#!/usr/bin/python2.4
import random
import gtk, gobject, cairo, rsvg, math

# Some flies
# Karl Lattimer <karl@qdh.org.uk>

class Fly:
    def __init__(self, num_of_flies, width=None, height=None, max_speed=None, x=None, y=None, velocity_x=None, velocity_y=None):
        if not x:
            self.x = random.randint(0, width)
        else:
            self.x = x
        if not y:
            self.y = random.randint(0, height)
        else:
            self.y = y
        if not velocity_x:
            self.velocity_x = random.randint(-1*max_speed, max_speed)
        else:
            self.velocity_x = velocity_x
        if not velocity_y:
            self.velocity_y = random.randint(-1*max_speed, max_speed)
        else:
            self.velocity_y = velocity_y
        
        self.num_of_flies = num_of_flies
        
        self.neighbours = []
        self.AddNeighbour()
        self.AddNeighbour()
        
    def AddNeighbour(self, fly=None):
        if not fly:
            fly = random.randint(0, self.num_of_flies-1)
        self.neighbours.append(fly)
    def SetNumOfFlies(self, num_of_flies):
        self.num_of_flies = num_of_flies
    def ReplaceNeighbour(self, old_fly, new_fly=None, num_of_flies=None ):
        if not new_fly:
            if num_of_flies:
                self.num_of_flies = num_of_flies
            new_fly = random.randint(0, self.num_of_flies-1)
        try:
            self.neighbours.remove(old_fly)
        except:
            pass
        self.neighbours.append(new_fly)
    
    def ReplaceRandom(self, num_of_flies=None, hit=None):
        if num_of_flies:
            self.num_of_flies = num_of_flies
        if not hit:
            hit = random.randint(0, len(self.neighbours))
        for index, neighbour in enumerate(self.neighbours):
            if index == hit:
                self.neighbours[index] = random.randint(0, self.num_of_flies-1)
    
    def Move( self, x=None, y=None ):
        if not x:
            x = self.x+self.velocity_x
        if not y:
            y = self.y+self.velocity_y
        self.tail_x = self.x
        self.tail_y = self.y
        self.x = x
        self.y = y
        
class Swarm:
    def __init__(self, flies, behavior, rect, color, cr):
        (impact_velocity, max_speed, grow_rate, decay_rate, type, size) = behavior
        (self.c_red, self.c_green, self.c_blue) = color
        (self.x, self.y, self.width, self.height) = rect
        self.fly_type = type
        self.cr = cr
        self.grow_rate = grow_rate
        self.decay_rate = decay_rate
        self.num_of_flies = flies
        self.flies = []
        self.impact_velocity = impact_velocity
        self.max_speed = max_speed
        self.stale_ticks = 100
        self.x_grav = 2
        self.y_grav = 2
        self.fresh_neighbour = self.stale_ticks
        self.repel_distance = 15
        self.flush = False
        self.feed_volume = 0
        self.size = size
        self.predator = None
        self.prey = None
        self.prey_volume = None
        self.type = type
        for fly in range(flies):
            self.flies.append( Fly(self.num_of_flies, self.width, self.height, self.max_speed) )

    def SetChain(self, predator, prey, prey_volume):
        self.predator = predator
        self.prey = prey
        self.prey_volume = prey_volume
        
    def FlushRefresh(self):
        global cr
        self.cr = cr
        self.cr.save()
        self.cr.set_source_rgb(.3, 0, .1)
        self.cr.rectangle(0, 0, self.width, self.height)
        self.cr.fill()
        self.cr.restore()
        self.cr.save()
        self.flush = True
        cr = self.cr
        self.Refresh()
        
    def Refresh(self):
        global cr
        self.cr = cr
        gen = False
        if len(self.flies) < 1:
            return
        idx = random.randint(0, self.num_of_flies-1)
        for self.fly,fly in enumerate(self.flies):
            hint = float(self.fly)/float(self.num_of_flies*2)
            self.fly = self.fly - 1
            self._proximity()
            self._evade()
            self._grav()
            self._hunt()
            # Give each fly and identity in colour
            self.co_red = self.c_red
            self.co_green = self.c_green
            self.co_blue = self.c_blue
            self.c_red = self.c_red + hint
            self.c_green = self.c_green + hint
            self.c_blue = self.c_blue + hint
            self._render()


            if self.fly == idx:
                gen = True
            else:
                gen = False

            if gen == True:
                if self.fresh_neighbour == 0:
                    self.fresh_neighbour = self.stale_ticks
                    self.flies[self.fly].ReplaceRandom(self.num_of_flies)
                else:
                    self.fresh_neighbour = self.fresh_neighbour - 1 
                
                if self.fresh_neighbour % self.grow_rate == 0:
                    self.num_of_flies = self.num_of_flies + 1
                    self.flies.append( Fly(self.num_of_flies, self.width, self.height, self.max_speed) )
                #    break;
                
                if (self.fresh_neighbour + 3) % self.decay_rate == 0:
                    self.num_of_flies = self.num_of_flies - 1
                    self.flies.remove(self.flies[0])
                #    self.flies = self.flies[0:self.fly]+self.flies[self.fly+1:self.num_of_flies-1]
                #    break;
            
            self.c_red = self.co_red
            self.c_green = self.co_green
            self.c_blue = self.co_blue

        
        
    def _grav(self):
        if self.flies[self.fly].x < self.width/3:
            self.flies[self.fly].velocity_x = self.flies[self.fly].velocity_x + random.randint(0, self.x_grav)
        if self.flies[self.fly].x > 2*(self.width/3):
            self.flies[self.fly].velocity_x = self.flies[self.fly].velocity_x - random.randint(0, self.x_grav)
        if self.flies[self.fly].y < self.height/3:
            self.flies[self.fly].velocity_y = self.flies[self.fly].velocity_y + random.randint(0, self.y_grav)
        if self.flies[self.fly].y > 2*(self.height/3):
            self.flies[self.fly].velocity_y = self.flies[self.fly].velocity_y - random.randint(0, self.y_grav)
                
    def _gen_birth(self):
        pass
    def _gen_death(self):
        pass
    def _evade(self):
        if self.flies[self.fly].x > (self.width - 20):
            self.flies[self.fly].velocity_x =  -1 * random.randint(0, self.impact_velocity)
        elif self.flies[self.fly].x < 20:
            self.flies[self.fly].velocity_x = random.randint(0, self.impact_velocity)

        if self.flies[self.fly].y > (self.height - 20):
            self.flies[self.fly].velocity_y = -1 *random.randint(0, self.impact_velocity)
        elif self.flies[self.fly].y < 20:
            self.flies[self.fly].velocity_y =  random.randint(0, self.impact_velocity)
        
    def _hunt(self):
        if not self.prey:
            return
        for idx,fly in enumerate(self.prey.flies):
            distance = self._distance(self.flies[self.fly], self.prey.flies[idx])
            if distance < 8:
                self.prey.num_of_flies = self.prey.num_of_flies - 1
                self.prey.flies.remove(self.prey.flies[idx])
                self.feed_volume = self.feed_volume + 1
                print self.type + " FEEDING"
            elif distance < 40: 
                dist_x = self.prey.flies[idx].x - self.flies[self.fly].x
                dist_y = self.prey.flies[idx].y - self.flies[self.fly].y
                
                if math.fabs(dist_x) > 40 or math.fabs(dist_y) > 40:
                    direction = 1
                else:
                    direction = -1
                    
                if dist_x != 0:
                    vx = self.flies[self.fly].velocity_x
                    self.flies[self.fly].velocity_x = vx + direction * (math.fabs(dist_x)/dist_x)
                vx = self.flies[self.fly].velocity_x
                if self.flies[self.fly].velocity_x > self.max_speed:
                    self.flies[self.fly].velocity_x = (math.fabs(vx)/2)
                if self.flies[self.fly].velocity_x < (-1*self.max_speed):
                    self.flies[self.fly].velocity_x = (-1*(math.fabs(vx)/2))
                if dist_y != 0:
                    vy = self.flies[self.fly].velocity_y
                    self.flies[self.fly].velocity_y = vy + direction * (math.fabs(dist_y)/dist_y)
                vy = self.flies[self.fly].velocity_y
                if self.flies[self.fly].velocity_y > self.max_speed:
                    self.flies[self.fly].velocity_y = (math.fabs(vx)/2)
                if self.flies[self.fly].velocity_y < (-1*self.max_speed):
                    self.flies[self.fly].velocity_y = (-1*(math.fabs(vx)/2))
                
        if self.feed_volume >= self.prey_volume:
            self.num_of_flies = self.num_of_flies + 1
            self.flies.append( Fly(self.num_of_flies, self.width, self.height, self.max_speed) )
            self.feed_volume = 0
            print self.type + " SPAWNING"
            
    def _proximity(self):
        # If the neighbours of this flys neighbours are closer then repatriate them
        # to this fly.
        for index, neighbour in enumerate(self.flies[self.fly].neighbours):
            if neighbour > self.num_of_flies - 1:
                self.flies[self.fly].ReplaceRandom(self.num_of_flies, index)
                continue
            distance = self._distance(self.flies[neighbour], self.flies[self.fly])
            for index, next_neighbour in enumerate(self.flies[neighbour].neighbours):
                if next_neighbour > self.num_of_flies - 1:
                    self.flies[self.fly].ReplaceRandom(self.num_of_flies, index)
                    continue
                next_distance  = self._distance(self.flies[next_neighbour], self.flies[self.fly]) 
                if next_distance > distance and next_neighbour != self.fly:
                    change = True
                    for t_neighbour in self.flies[self.fly].neighbours:
                        if t_neighbour == next_neighbour:
                            change = False
                    if change:
                        self.flies[self.fly].ReplaceNeighbour(neighbour, next_neighbour)
                        
            dist_x = self.flies[neighbour].x - self.flies[self.fly].x
            dist_y = self.flies[neighbour].y - self.flies[self.fly].y
            
            if math.fabs(dist_x) < self.repel_distance or math.fabs(dist_y) < self.repel_distance:
                direction = -1
            else:
                direction = 1
            if dist_x != 0:
                vx = self.flies[self.fly].velocity_x
                self.flies[self.fly].velocity_x = vx + direction * (math.fabs(dist_x)/dist_x)
            vx = self.flies[self.fly].velocity_x
            if self.flies[self.fly].velocity_x > self.max_speed:
                self.flies[self.fly].velocity_x = (math.fabs(vx)/2)
            if self.flies[self.fly].velocity_x < (-1*self.max_speed):
                self.flies[self.fly].velocity_x = (-1*(math.fabs(vx)/2))
            if dist_y != 0:
                vy = self.flies[self.fly].velocity_y
                self.flies[self.fly].velocity_y = vy + direction * (math.fabs(dist_y)/dist_y)
            vy = self.flies[self.fly].velocity_y
            if self.flies[self.fly].velocity_y > self.max_speed:
                self.flies[self.fly].velocity_y = (math.fabs(vx)/2)
            if self.flies[self.fly].velocity_y < (-1*self.max_speed):
                self.flies[self.fly].velocity_y = (-1*(math.fabs(vx)/2))
            self.flies[self.fly].Move()
            
    def _collission(self):
        pass
    def _distance( self, a, b ):
        # I'm sure pythagoras is released under GPL
        x = a.x - b.x
        y = a.y - b.y
        return int(math.hypot(x, y))
    
    def _rtop(self, x, y):
        length = math.hypot(x, y)
        if y != 0:
            angle = math.atan(x/y)
        else:
            angle = 0
        return (angle, length)
    
    def _ptor(self, angle, length):
        #soh cah toa
        y = length * math.cos(angle)
        x = length * math.sin(angle)
        return (x, y)
    
    def _render(self):
        self.cr.set_source_rgb(self.c_red, self.c_green, self.c_blue)
        self.cr.arc(self.flies[self.fly].x, self.flies[self.fly].y, self.size, 0, 2.0 * math.pi)
        self.cr.set_line_width( 2*(self.size/3) )
        self.cr.fill()
        
        try:
            (angle, length) = self._rtop(self.flies[self.fly].x - self.flies[self.fly].tail_x,
                                         self.flies[self.fly].y - self.flies[self.fly].tail_y)
        except:
            (angle, length) = self._rtop(self.flies[self.fly].x - self.flies[self.fly].x,
                                         self.flies[self.fly].y - self.flies[self.fly].y)
            self.flies[self.fly].tail_x = self.flies[self.fly].x
            self.flies[self.fly].tail_y = self.flies[self.fly].y
        (tail_x, tail_y) = (self.flies[self.fly].tail_x, self.flies[self.fly].tail_y)
        
        self.cr.move_to( self.flies[self.fly].x, self.flies[self.fly].y )
        self.cr.line_to( tail_x, tail_y )
        self.cr.stroke()
        
class FliesWidget(gtk.DrawingArea):
    __gsignals__ = { "expose-event": "override" }
    
    def do_expose_event( self, event ):
        global swarma
        global swarmb
        global cr
    
        self.width = event.area.width
        self.height = event.area.height
        self.cr = self.window.cairo_create()
        self.MyExpose()
        
    def MyExpose(self):
        global swarma
        global swarmb
        global swarmc
        global cr
        
        behavior = (8, 18, 9, 7, "red", 5)
        color = (.7, 0, .3)
        rect = (0, 0, self.width, self.height)
        if not swarma:
             swarma = Swarm(2, behavior, rect, color, self.cr )
        else:
            swarma.width =  self.width
            swarma.height = self.height
            swarma.cr = self.cr
        
        swarma.FlushRefresh()
        
        behavior = (8, 17, 9, 8, "blue", 3)
        color = (.1, .3, .6)
        rect = (0, 0, self.width, self.height)
        if not swarmb:
             swarmb = Swarm(3, behavior, rect, color, self.cr )
        else:
            swarmb.width = self.width
            swarmb.height = self.height
            swarmb.cr = self.cr
            
        swarmb.Refresh()
        
        behavior = (6, 12, 6, 12, "green", 2)
        color = (.1, .7, .2)
        rect = (0, 0, self.width, self.height)
        if not swarmc:
             swarmc = Swarm(10, behavior, rect, color, self.cr )
        else:
            swarmc.width = self.width
            swarmc.height = self.height
            swarmc.cr = self.cr
        swarma.SetChain( None, swarmb, 2 )
        swarmb.SetChain( swarma, swarmc, 2 )
        swarmc.SetChain( swarmb, None, None )
        swarmc.Refresh()
        reds = "Reds %d " % swarma.num_of_flies
        blues = "Blues %d " % swarmb.num_of_flies
        greens = "Greens %d" % swarmc.num_of_flies
        print reds + blues + greens
        gobject.timeout_add(1000/20, self.MyExpose)
        
if __name__ == "__main__":
    print "Starting"
    swarma = None
    swarmb = None
    swarmc = None
    window = gtk.Window()
    window.connect("delete-event", gtk.main_quit)
    dc = FliesWidget()
    
    window.resize(1000, 700)
    dc.show()
    window.add(dc)
    
    window.present()
    cr = dc.window.cairo_create()
    gtk.main()